import { Box } from '../../classes/display/box.js';
import { Scene } from '../../classes/display/scene.js';
import { PostContent } from '../../classes/supporting/post-content.js';
import { Expert } from '../../classes/actors/expert.js';
import { ForumNode } from '../../classes/forum-network/forum-node.js';
import { Network } from '../../classes/forum-network/network.js';
import { mochaRun, randomID } from '../../util/helpers.js';
import { delayOrWait } from '../../classes/display/controls.js';

describe('Forum Network', function tests() {
  this.timeout(0);

  let scene;
  let author1;
  let author2;
  let forumNetwork;
  let forumNode1;
  let forumNode2;
  let forumNode3;
  let processInterval;

  before(async () => {
    const rootElement = document.getElementById('scene');
    const rootBox = new Box('rootBox', rootElement).flex();

    scene = new Scene('Forum Network test', rootBox).withSequenceDiagram();

    author1 = await new Expert(null, 'author1', scene).initialize();
    author2 = await new Expert(null, 'author2', scene).initialize();

    forumNetwork = new Network();

    forumNode1 = await new ForumNode('node1', scene).initialize(
      forumNetwork,
    );
    forumNode2 = await new ForumNode('node2', scene).initialize(
      forumNetwork,
    );
    forumNode3 = await new ForumNode('node3', scene).initialize(
      forumNetwork,
    );

    processInterval = setInterval(async () => {
      await forumNode1.processNextMessage();
      await forumNode2.processNextMessage();
      await forumNode3.processNextMessage();
    }, 100);
  });

  after(() => {
    clearInterval(processInterval);
  });

  // const blockchain = new Blockchain();

  specify('Author can submit a post to the network', async () => {
    const post1 = new PostContent({ message: 'hi' });
    post1.id = randomID();
    const post2 = new PostContent({ message: 'hello' }).addCitation(
      post1.id,
      1.0,
    );

    await delayOrWait(1000);
    await author1.submitPostViaNetwork(
      forumNode1,
      post1,
      50,
    );
    await delayOrWait(1000);
    await author2.submitPostViaNetwork(
      forumNode2,
      post2,
      100,
    );

    await delayOrWait(1000);
  }).timeout(10000);
});

mochaRun();
